#include <time.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include "leafnode.h"

int debug = 0;

struct newsgroup * active;

/* this uses a nifty technique for building a binary tree from a
   sorted list... the basic observation is that if you count from one
   end, with 1 being the rightmost node, then the number of nodes
   between N and the leaf nodes is equal to the number of zero bits
   at the small end of N.  this tree probably doesn't make it obvious:

   1 2 3 4 5 6 7

         1
     1   0   1
   1 0 1 0 1 0 1
   | | | | | | + no zeroes: leaf node
   | | | | | +-- one zero: one step up
   | | | | +---- no zeroes: leaf node
   | | | +------ two zeroes: two steps up
   | | +-------- no zeroes: leaf node
   | +---------- one zero: one step up
   +------------ no zeroes: leaf node

   this can be used to build a smallest-height tree, and with a bit more
   care, a _perfectly_ balanced tree */

void readactive( void ) {
    char * p, *q, *r, *lastline, *desc;
    int first, last;
    time_t age;
    FILE * f;
    struct newsgroup * g;
    struct stat st;

    int line;
    struct newsgroup *levels[32]; /* theoretically finite size :) */
    int l, n, error;

    char * stuff;

    active = NULL;

    strcpy(s, spooldir);
    strcat(s, "/leaf.node/groupinfo");
    if ( stat( s, &st ) ) {
	syslog( LOG_ERR, "can't stat %s: %m", s );
	return;
    }
    stuff = critmalloc( st.st_size+2, "Reading group info" );
    if (( f = fopen( s, "r" )) != NULL ) {
	n = fread( stuff, 1, st.st_size, f );
	if ( n < st.st_size )
	    syslog( LOG_ERR,
		    "Groupinfo file truncated while reading: %d < %d .",
		    n, st.st_size );
	fclose( f );
    }
    else {
	syslog( LOG_ERR, "unable to open %s: %m", s );
	return;
    }

    stuff[st.st_size] = '\n';
    stuff[st.st_size+1] = '\0';	/* 0-terminate string */

    printf("readactive1: %d\n", time(NULL) );

    /* delete spurious 0-bytes */
    for ( l = 0; l < st.st_size; l++ )
	if ( stuff[l] == '\0' )
	    stuff[l] = ' ';  /* \n might be better, but produces more errors */

    line = 1;
    for( l=0; l<32; levels[l++] = NULL )
	;

    printf("readactive2: %d\n", time(NULL) );

    p = stuff;
    error = 0;
    while ( (p < (stuff+n)) && (( q = strchr( p, '\n' )) != NULL ) ) {
	*q = '\0' ;
	lastline = p;
        if (( ( r = strchr( p, ' ' )) != NULL ) && isalnum( *p ) ) {
            *r = '\0' ;
            p = r + 1 ;
        } else
            error = 1;
        if ( !error ) {
            last = p ? strtol( p, &r, 10 ) : 0;
            if ( r == p )
                error = 1;
            else {
                p = r;
                first = p ? strtol( p, &r, 10 ) : 1;
                if ( r == p )
                    error = 1;
                else {
                    p = r;
                    age = p ? strtol( p, &r, 10 ) : 1;
                    if ( r == p )
                        error = 1;
		    else
			p = r;
                }
            }
        }

        if ( !error ) {
            while (p && *p && isspace(*p))
                p++;
            if ( p && *p )
                desc = strdup( p );
            else
                error = 1;
        }

        if ( error ) {
            syslog( LOG_ERR,
		    "Groupinfo file possibly truncated or damaged: %s\n",
		    lastline);
            error = 0;
        } else {
	    g = (struct newsgroup *) critmalloc(sizeof(struct newsgroup),
						"reading groupinfo" );
	    g->right = g->left = NULL;
	    g->first = first;
	    g->last = last;
	    g->age = age;
	    g->name = strdup(lastline);
	    g->desc = desc;

            /* Inserting newsgroup into tree */
            for( l=0; (line&(1<<l))==0; l++ )
                ;
            if ( l ) {
                g->right = levels[l-1];
            }
            if ( active == g->right )
                active = g;
            if ( l<31 && levels[l+1] && !levels[l+1]->left )
                levels[l+1]->left = g;
            levels[l] = g;

            p = strchr( p, '\n' );
            if ( p )
                *p++ = '\0';
            line++;
	}

        p = q + 1;
    }
    free( stuff );

    printf("readactive3: %d\n", time(NULL) );

    /* More tree stuff */
    g = NULL;
    for ( l=0; l<31; l++ ) {
	if ( levels[l] ) {
	    if ( g && levels[l] &&
		 levels[l]->left == NULL && levels[l]->right != g ) {
		levels[l]->left = g;
		g = NULL;
	    }
	    if ( !levels[l+1] ||
		 ( levels[l] != levels[l+1]->left &&
		   levels[l] != levels[l+1]->right ) ) {
		if ( g )
		    syslog( LOG_ERR, "2+5=2" );
		g = levels[l];
	    }
	}
    }
}

void main ( void ) {
    struct stat st;
    char * stuff;
    FILE * f;
    int l, n;

    printf("Vor readconfig: %d\n", time(NULL) );
    readconfig();
    printf("Vor readactive: %d\n", time(NULL) );
    readactive();
    printf("    readactive: %d\n", time(NULL) );

}
